package log;

import nl.west.rme.common.Env;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.logging.Handler;
import java.util.logging.Level;
import java.util.logging.LogRecord;


public class Log
{
  private static final java.util.logging.Logger unsafe =
    java.util.logging.Logger.getLogger("Log");
  private static final List<Handler> classloader_local_handlers = Collections
      .synchronizedList(new ArrayList<Handler>());

  private static final ThreadLocal<Boolean> local_log_enabled;
  private static final ThreadLocal<List<String>> local_log_history;
  private static final ThreadLocal<Long> last_log_timestamp;

  public static volatile boolean report_large_log_intervals = false;
  public static volatile long report_large_log_interval_ms = 250L;

  static {
    unsafe.setLevel(Level.ALL);
    unsafe.setFilter(null);

    local_log_enabled = new ThreadLocal<Boolean>()
    {
      @Override
      protected Boolean initialValue()
      {
        return Boolean.FALSE;
      }
    };

    local_log_history = new ThreadLocal<List<String>>()
    {
      @Override
      protected List<String> initialValue()
      {
        return new ArrayList<String>();
      }
    };

    last_log_timestamp = new ThreadLocal<Long>()
    {
      @Override
      protected Long initialValue()
      {
        return Long.valueOf(System.currentTimeMillis());
      }
    };
  }

  public static void addHandler(Handler handler)
  {
    // logger.addHandler(handler);
    // the logger reference points to a VM-global instance,
    // regardless of ClassLoaders and web-contexts.
    //
    // we can have any amount of Log classes loaded, but
    // all of them will have Log.logger point to the same instance.
    //
    // once we add handlers to them, the handlers will receive
    // log-records from classloaders / web-contexts they shouldn't
    // have access to, allowing web-contexts to mess with
    // eachothers' data and/or state.

    classloader_local_handlers.add(handler);
    System.out.println("Log.addHandler: #" + classloader_local_handlers.size());
  }

  public static void signalNewTask()
  {
    last_log_timestamp.set(Long.valueOf(System.currentTimeMillis()));
  }

  public static void enableHistory()
  {
    local_log_enabled.set(Boolean.TRUE);
  }

  public static List<String> disableHistory()
  {
    local_log_enabled.set(Boolean.FALSE);

    final List<String> history = local_log_history.get();
    local_log_history.set(new ArrayList<String>());
    return history;
  }

  public static void custom(Level level, String msg)
  {
    log(level, msg, null, false);
  }

  public static void debug(String msg)
  {
    log(Level.FINE, msg, null, false);
  }

  public static void debug(String msg, Throwable problem)
  {
    log(Level.FINE, msg, problem, false);
  }

  public static void info(String msg)
  {
    log(Level.INFO, msg, null, false);
  }

  public static void info(String msg, Throwable problem)
  {
    log(Level.INFO, msg, problem, false);
  }

  public static void warn(String msg)
  {
    log(Level.WARNING, msg, null, false);
  }

  public static void warn(String msg, Throwable problem)
  {
    log(Level.WARNING, msg, problem, false);
  }

  public static void error(String msg)
  {
    log(Level.SEVERE, msg, null, false);
  }

  public static void error(String msg, Throwable problem)
  {
    log(Level.SEVERE, msg, problem, false);
  }

  public static void error(String msg, Throwable problem, boolean sendMail)
  {
    log(Level.SEVERE, msg, problem, sendMail);
  }

  private static void log(
      Level level, String msg, Throwable problem, boolean sendMail)
  {
    if (report_large_log_intervals) {
      final long now = System.currentTimeMillis();
      final long elapsed = now - last_log_timestamp.get().longValue();
      if (elapsed > report_large_log_interval_ms)
        logImpl(Level.FINE, "HICKUP: " + elapsed + "ms", true);
    }

    {
      // a sane log-viewer would use LogRecord.seqNr to sort records,
      // instead of a millisecond precision timestamp...
      // in google's log-viewer records with an equal timestamp are
      // arbitrarily ordered. force the timestamps to be increasing:

      while (last_log_timestamp.get() == System.currentTimeMillis())
        continue; // ensure log entries (sorted by time) retain their order -
                  // N.B.: do not sleep!
      last_log_timestamp.set(Long.valueOf(System.currentTimeMillis()));
    }

    msg = trace() + ": " + msg;

    if (Env.isLocal()) {
      final boolean isErr = (level.intValue() >= Level.WARNING.intValue());
      final PrintStream str = isErr ? System.err : System.out;
      str.println(level.getName() + " " + msg);
      if (problem != null)
        problem.printStackTrace(str);

      logImpl(level, msg, false);
      return;
    }

    try {
      if (local_log_enabled.get()) {
        local_log_history
            .get().add(level.getName().substring(0, 4) + " " + msg);
      }
    }
    catch (final Exception exc) {
      exc.printStackTrace();
    }

    logImpl(level, msg, true);

    String stacktrace = "No stacktrace";
    if (problem != null) {
      final StringWriter sw = new StringWriter();
      problem.printStackTrace(new PrintWriter(sw));
      stacktrace = sw.getBuffer().toString();
      logImpl(level, stacktrace, true);
    }

    if (sendMail) {
      // try {
      // final EmailMessage email = new EmailMessage.Builder(msg, stacktrace)//
      // .setMimeType(EmailMessage.MimeType.TEXT)//
      // .build();
      // email.setFrom(AppspotEmail.getAddress("noreply"), LOG_NAME)//
      // .addTo(LOG_EMAIL, LOG_NAME);
      // final EmailService service = new EmailService(email);
      // service.send();
      // }
      // catch (final Exception e) {
      // Log.error("Error in sending log error.", e);
      // }
    }
  }

  private static final void logImpl(
      Level level, String msg, boolean notifyLogger)
  {
    if (notifyLogger) {
      unsafe.log(level, msg);
    }

    // do NOT use logger.addHandler
    synchronized (classloader_local_handlers) {
      if (!classloader_local_handlers.isEmpty()) {
        final LogRecord record = new LogRecord(level, msg);

        for (final Handler handler : classloader_local_handlers) {
          handler.publish(record);
        }
      }
    }
  }

  private static String trace()
  {
    boolean next = false;
    for (final StackTraceElement elem : new Throwable().getStackTrace()) {
      if (elem.getClassName().equals(Log.class.getName())) {
        next = true;
      }
      else if (next) {
        String clazz = elem.getClassName();
        int io = clazz.lastIndexOf('.');
        if (io != -1)
          clazz = clazz.substring(io + 1);

        io = clazz.lastIndexOf('$');
        if (io != -1 && clazz.length() - io > 3)
          clazz = clazz.substring(io + 1);

        final String lineNumber = String.valueOf(elem.getLineNumber());
        final String methodName = elem.getMethodName();
        return clazz + (methodName.startsWith("<") ? "" : ("." + methodName))
            + "(" + lineNumber + ")";
      }
    }

    return "[no trace]";
  }
}
