package net.indiespot.sql.data;

import log.Log;
import nl.west.rme.common.util.CollectionUtils;
import nl.west.rme.common.util.GenericsUtils;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;


public class Fetcher
{
  public boolean isDone()
  {
    return type2fetchKeys.isEmpty();
  }

  public void clear()
  {
    type2fetchKeys.clear();
    type2key2item.clear();
  }

  public Fetcher deepCopy()
  {
    final Fetcher copy = new Fetcher();

    for (final Class<?> type : type2key2item.keySet()) {
      final Map<Key<?>, Record<?>> key2item = new HashMap<>();
      copy.type2key2item.put(type, key2item);

      for (final Record<?> item : type2key2item.get(type).values()) {
        final Record<?> item2 = item.copy();
        key2item.put(item2.getKey(), item2);
      }
    }

    return copy;
  }

  private final Map<Class<?>, Set<Key<?>>> type2fetchKeys = new HashMap<>();

  public <T extends Record<T>> Key<T> addKey(Key<T> key)
  {
    // discard the keys we already have
    final Map<Key<?>, Record<?>> key2item = type2key2item.get(key.type);
    if (key2item != null && key2item.containsKey(key))
      return key;

    // add these new key
    Set<Key<?>> typedKeys = type2fetchKeys.get(key.type);
    if (typedKeys == null) {
      final Set<Key<T>> keySet = createTypedKeySet(key.type);
      typedKeys = GenericsUtils.cast(keySet);
      type2fetchKeys.put(key.type, typedKeys);
    }
    typedKeys.add(key);

    return key;
  }

  public <T extends Record<T>> void addKeys(Set<Key<T>> keys)
  {
    for (final Key<T> key : keys) {
      this.addKey(key);
    }
  }

  public <T extends Record<T>> Set<T> getKeys(Class<T> type)
  {
    final Set<Key<?>> keys = new HashSet<>();

    final Set<Key<?>> fetchKeys = type2fetchKeys.get(type);
    if (fetchKeys != null)
      keys.addAll(fetchKeys);

    final Map<Key<?>, Record<?>> key2item = type2key2item.get(type);
    if (key2item != null)
      keys.addAll(key2item.keySet());

    return GenericsUtils.cast(keys);
  }

  public <T extends Record<T>> void removeKey(Key<T> entityKey)
  {
    final Set<Key<?>> fetchKeys = type2fetchKeys.get(entityKey.type);
    if (fetchKeys != null)
      fetchKeys.remove(entityKey);

    final Map<Key<?>, Record<?>> key2item = type2key2item.get(entityKey.type);
    if (key2item != null)
      key2item.remove(entityKey);
  }

  @SuppressWarnings("unchecked")
  public Set<Class<Record<?>>> getEntityTypes()
  {
    final Set<Class<Record<?>>> types = new HashSet<>();
    for (final Class<?> type : type2fetchKeys.keySet())
      types.add((Class<Record<?>>) type);
    for (final Class<?> type : type2key2item.keySet())
      types.add((Class<Record<?>>) type);
    return types;
  }

  public <T extends Record<T>> T getEntity(Key<T> key)
  {
    if (key == null)
      return null;

    final T entity = this.resolve(key);

    if (entity == null) {
      this.addKey(key);
      this.fetch();
    }
    return this.resolve(key);
  }

  public <T extends Record<T>> Set<T> getEntities(Class<T> type)
  {
    final Set<Key<T>> fetchKeys = GenericsUtils.cast(type2fetchKeys.get(type));
    if (fetchKeys == null || fetchKeys.isEmpty()) {
      final Map<Key<?>, Record<?>> key2item = type2key2item.get(type);
      if (key2item == null)
        return new HashSet<>();
    }
    else {
      this.fetch();
    }

    final Collection<T> items =
      GenericsUtils.cast(type2key2item.get(type).values());
    return new HashSet<T>(items);
  }

  public <T extends Record<T>> Set<T> fetchPendingEntities(Class<T> type)
  {
    final Set<Key<T>> fetchKeys = GenericsUtils.cast(type2fetchKeys.get(type));
    if (fetchKeys == null || fetchKeys.isEmpty())
      return new HashSet<>();

    return this.getEntities(new HashSet<Key<T>>(fetchKeys));
  }

  @SuppressWarnings("unchecked")
  public <T extends Record<T>> Set<T> getEntities(Set<Key<T>> keys)
  {
    if (keys.isEmpty())
      return new HashSet<>();

    this.addKeys(keys);

    // ensure we have all keys fetched
    final Class<?> type = keys.iterator().next().type;
    Map<Key<?>, Record<?>> key2item = type2key2item.get(type);
    if (key2item == null || !key2item.keySet().containsAll(keys))
      this.fetch(); // pull in pending entities
    key2item = type2key2item.get(type);

    // resolve resulting entities
    final Set<T> key2result = new HashSet<>();
    for (final Key<T> key : keys) {
      final T resolvedTo = (T) key2item.get(key);
      if (resolvedTo == null)
        key.flagMissing();
      else
        key.setResolvedTo(resolvedTo, this);
      key2result.add(resolvedTo);
    }
    return key2result;
  }

  public <T extends Record<T>> boolean hasEntity(T entity)
  {
    return this.hasEntityWithKey(entity.getKey());
  }

  public <T extends Record<T>> boolean hasEntityWithKey(Key<T> entityKey)
  {
    final Map<Key<?>, Record<?>> map = type2key2item.get(entityKey.type);
    return (map != null) && map.containsKey(entityKey);
  }

  public <T extends Record<T>> void removeEntity(T entity)
  {
    this.removeKey(entity.getKey());
  }

  public <T extends Record<T>> void removeEntities(Collection<T> entities)
  {
    for (final T entity : entities) {
      this.removeEntity(entity);
    }
  }

  public <T extends Record<T>> void removeEntities(Class<T> type)
  {
    type2fetchKeys.remove(type);
    type2key2item.remove(type);
  }

  public <T extends Record<T>> Set<T> drainEntities(Class<T> type)
  {
    final Set<T> set = this.getEntities(type);
    type2key2item.remove(type);
    return set;
  }

  public boolean verbose = false;

  @SuppressWarnings("unchecked")
  public <T extends Record<T>> Fetcher fetch()
  {
    if (type2fetchKeys.isEmpty())
      return this;

    for (final Entry<Class<?>, Set<Key<?>>> entry : type2fetchKeys.entrySet()) {
      final Class<T> type = (Class<T>) entry.getKey();
      final Set<Key<T>> keys = GenericsUtils.cast(entry.getValue());

      // store the results
      Map<Key<?>, Record<?>> key2item = type2key2item.get(type);
      if (key2item == null)
        type2key2item.put(
            type, key2item = GenericsUtils.cast(createTypedKeyMap(type)));

      final Set<T> resultItems = JdbcRecordManager.queryByKeys(keys);
      if (verbose)
        Log.debug("queried " + keys.size() + " keys => " + resultItems.size()
            + " " + type.getSimpleName().toLowerCase() + " entities");
      for (final T item : resultItems)
        key2item.put(item.getKey(), item);
    }

    type2fetchKeys.clear();

    return this;
  }

  // -----------

  private final Map<Class<?>, Map<Key<?>, Record<?>>> type2key2item =
    new HashMap<>();

  public void addAll(Fetcher that)
  {
    this.addAllImpl(that);
  }

  @SuppressWarnings("unchecked")
  private <T extends Record<T>> void addAllImpl(Fetcher that)
  {
    for (final Entry<Class<?>, Map<Key<?>, Record<?>>> entry1 : that.type2key2item
        .entrySet()) {
      for (final Entry<Key<?>, Record<?>> entry2 : entry1.getValue().entrySet()) {
        this.add((T) entry2.getValue());
      }
    }
  }

  public <T extends Record<T>, C extends Collection<T>> C addAll(C items)
  {
    final C uniques = CollectionUtils.newCollectionByType(items);
    for (final T item : items)
      uniques.add(this.add(item));
    return uniques;
  }

  @SuppressWarnings({ "unchecked", "rawtypes" })
  public <T extends Record<T>> T add(T item)
  {
    if (item == null)
      throw new NullPointerException();

    final Key<T> key = item.getKey();
    final Class<T> type = key.type;

    // remove from pending keys
    final Set<Key<?>> pendingKeys = this.type2fetchKeys.get(type);
    if (pendingKeys != null)
      pendingKeys.remove(key);

    // add to resolved keys
    Map key2item = type2key2item.get(type);
    if (key2item == null)
      type2key2item.put(type, key2item = createTypedKeyMap(type));

    final Record<?> old = (Record<?>) key2item.put(key, item);
    if (old != null && old != item) {
      // FIXME Log.debug("duplicate entity: " + item);
      key2item.put(key, old); // restore old value in map
      return (T) old;
    }

    return item;
  }

  @SuppressWarnings({ "rawtypes", "unchecked" })
  public <T extends Record<T>> T resolve(Key<T> key)
  {
    if (key == null)
      return null;
    if (key.isResolvedWith(this))
      return key.value();

    final Map key2item = type2key2item.get(key.type);
    if (key2item == null)
      return null;
    final T item = (T) key2item.get(key);
    if (item != null || key2item.containsKey(key))
      key.setResolvedTo(item, this);
    return item;
  }

  public Fetcher unresolveForeignKeys()
  {
    final long t0 = System.nanoTime();

    for (final Entry<Class<?>, Map<Key<?>, Record<?>>> entry : type2key2item
        .entrySet())
      for (final Record<?> record : entry.getValue().values())
        record.unresolveKeys();

    final long t1 = System.nanoTime();
    Log.debug("took: " + (t1 - t0) / 1_000_000L + "ms");

    return this;
  }

  public Fetcher resolve()
  {
    this.fetch();
    this.resolveOutboundForeignKeys();
    this.resolveInboundForeignKeys();
    return this;
  }

  private Fetcher resolveOutboundForeignKeys()
  {
    for (final Entry<Class<?>, Map<Key<?>, Record<?>>> entry : type2key2item
        .entrySet())
      for (final Record<?> record : entry.getValue().values())
        record.resolveKeys(this);
    return this;
  }

  private void resolveInboundForeignKeys()
  {
    final List<RecordManagerDomain> ctxs = RecordManagerDomain.Access.get();
    for (int i = 0, len = ctxs.size(); i < len; i++)
      ctxs.get(i).resolveInboundForeignKeys(this);
  }

  public Fetcher reset()
  {
    type2key2item.clear();

    return this;
  }

  public int size()
  {
    int sum = 0;
    for (final Entry<Class<?>, Map<Key<?>, Record<?>>> entry1 : type2key2item
        .entrySet())
      sum += entry1.getValue().size();
    return sum;
  }

  @SuppressWarnings("serial")
  private static <T extends Record<T>> Set<Key<T>> createTypedKeySet(
      final Class<T> type)
  {
    return new HashSet<Key<T>>()
    {
      @Override
      public boolean add(Key<T> key)
      {
        if (key.type != type)
          throw new IllegalStateException("key type mismatch: key.type="
              + key.type + " vs required=" + type);
        if (key.id == 0)
          throw new IllegalStateException("zero key");
        // if(key.id < 0)
        // throw new IllegalStateException("negative key: " + key.id);
        return super.add(key);
      }
    };
  }

  @SuppressWarnings("serial")
  private static <T extends Record<T>> Map<Key<T>, T> createTypedKeyMap(
      final Class<T> type)
  {
    return new HashMap<Key<T>, T>()
    {
      @Override
      public T put(Key<T> key, T entity)
      {
        if (key.type != type)
          throw new IllegalStateException("key type mismatch: key.type="
              + key.type + " vs required=" + type);
        if (entity != null && entity.getClass() != type)
          throw new IllegalStateException("entity type mismatch: entity.type="
              + entity.getClass() + " vs required=" + type);
        if (key.id == 0)
          throw new IllegalStateException("zero key");
        // if(key.id < 0)
        // throw new IllegalStateException("negative key: " + key.id);
        if (entity != null && entity.id != key.id)
          throw new IllegalStateException("key/entity mismatch: key=" + key.id
              + " vs entity=" + entity.id);
        return super.put(key, entity);
      }
    };
  }
}
