package net.indiespot.sql.data;

import java.util.NoSuchElementException;

public final class Key<T extends Record<T>> {
	public static <T extends Record<T>> Key<T> create(Class<T> type, int id) {
		if (id == 0)
			return null;
		return new Key<T>(type, id);
	}

	public final Class<T> type;
	public final int id;

	public Key(Class<T> type, int id) {
		if (type == null)
			throw new IllegalStateException();
		if (id == 0)
			throw new IllegalStateException("id: " + id);
		this.type = type;
		this.id = id;
	}

	public static <Q extends Record<Q>> Key<Q> copy(Key<Q> key) {
		return key == null ? null : key.copy();
	}

	public Key<T> copy() {
		return Key.create(type, id);
	}

	public T ensure() {
		return this.isResolved() ? this.resolvedTo : this.select();
	}

	public T select() {
		if (id <= 0)
			throw new IllegalStateException();
		final T result = JdbcRecordManager.queryByKey(this);
		if (result == null)
			throw new NoSuchElementException(type.getSimpleName() + "#" + id);
		return this.setResolvedTo(result, null);
	}

	public T selectOrNull() {
		if (id <= 0)
			throw new IllegalStateException();
		final T result = JdbcRecordManager.queryByKey(this);
		if (result == null)
			return null;
		return this.setResolvedTo(result, null);
	}

	private static final Fetcher DUMMY = new Fetcher();
	private Fetcher resolvedWith;
	private boolean isMissing;
	private T resolvedTo;

	public boolean isResolved() {
		return resolvedWith != null;
	}

	public boolean isResolvedWith(Fetcher with) {
		if (with == null)
			throw new NullPointerException();
		return resolvedWith == with;
	}

	public T setResolvedTo(T value, Fetcher with) {
		if (value == null)
			throw new IllegalStateException("resolved to null: " + this);

		resolvedWith = (with == null ? DUMMY : with);
		resolvedTo = value;

		if (value.id != this.id)
			throw new IllegalStateException(
					"revolving with invalid key [got:" + value.getKey() + ", expected:" + this + "]");
		return value;
	}

	public T resolveWith(Fetcher fetcher) {
		final T got = fetcher.getEntity(this);
		if (got == null)
			this.flagMissing();
		else
			this.setResolvedTo(got, fetcher);
		return got;
	}

	public void unresolve() {
		resolvedWith = null;
		resolvedTo = null;
	}

	public void flagMissing() {
		resolvedWith = DUMMY;
		isMissing = true;
	}

	public T value() {
		if (!this.isResolved())
			throw new IllegalStateException("unresolved key: " + this);
		if (isMissing)
			throw new NoSuchElementException("corrupt foreign key");
		return resolvedTo;
	}

	public <R extends Record<R>> T value(R holder) {
		if (!this.isResolved())
			throw new IllegalStateException(
					"unresolved key: " + this + ", holder=" + (holder == null ? "null" : holder.toDetailedString()));
		if (isMissing)
			throw new NoSuchElementException(
					"corrupt foreign key, holder=" + (holder == null ? "null" : holder.toDetailedString()));
		return resolvedTo;
	}

	@Override
	public int hashCode() {
		return type.hashCode() ^ id;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;

		if (obj.getClass() != Key.class)
			throw new IllegalStateException();

		@SuppressWarnings("unchecked")
		final Key<T> that = (Key<T>) obj;
		if (type != that.type)
			return false;
		if (id != that.id)
			return false;

		return true;
	}

	@Override
	public String toString() {
		return "Key[" + type.getSimpleName() + "#" + id + (this.isResolved() ? "" : "[unresolved]") + "]";
	}
}
