package net.indiespot.sql.data;

import nl.west.rme.common.util.Compare;
import java.lang.reflect.Array;
import java.lang.reflect.Field;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.List;
import java.util.NoSuchElementException;


public abstract class Record<T extends Record<T>> implements Comparable<T>
{
  public int id;

  @SuppressWarnings({ "unchecked", "rawtypes" })
  public Key<T> getKey()
  {
    if (id == 0)
      throw new NoSuchElementException("key to "
          + this.getClass().getSimpleName());

    final Key<T> key = new Key(this.getClass(), id);
    key.setResolvedTo((T) this, null);
    return key;
  }

  protected void resolveKeys(Fetcher fetcher)
  {
    @SuppressWarnings("unchecked")
    final List<Field> fields = RecordTypeMetadata.getKeyFields(this.getClass());
    for (final Field field : fields) {
      try {
        fetcher.resolve((Key<?>) field.get(this));
      }
      catch (IllegalArgumentException | IllegalAccessException e) {
        throw new IllegalStateException(e);
      }
    }
  }

  protected void unresolveKeys()
  {
    @SuppressWarnings("unchecked")
    final List<Field> fields = RecordTypeMetadata.getKeyFields(this.getClass());
    for (final Field field : fields) {
      try {
        final Key<?> key = (Key<?>) field.get(this);
        if (key != null)
          key.unresolve();
      }
      catch (IllegalArgumentException | IllegalAccessException e) {
        throw new IllegalStateException(e);
      }
    }
  }

  public <X extends Record<X>> void insert()
  {
    @SuppressWarnings("unchecked")
    final X self = (X) this;
    JdbcRecordManager.insert(self);
  }

  public <X extends Record<X>> void update()
  {
    @SuppressWarnings("unchecked")
    final X self = (X) this;
    JdbcRecordManager.update(self);
  }

  public <X extends Record<X>> void delete()
  {
    @SuppressWarnings("unchecked")
    final X self = (X) this;
    JdbcRecordManager.delete(self);
  }

  public <X extends Record<X>> void save()
  {
    @SuppressWarnings("unchecked")
    final X self = (X) this;
    JdbcRecordManager.save(self);
  }

  @SuppressWarnings("unchecked")
  public T copy()
  {
    T copy;

    try {
      copy = RecordManager.newRecord((Class<T>) this.getClass());
    }
    catch (final Exception exc) {
      throw new IllegalStateException(exc);
    }

    final List<Field> fields =
      RecordTypeMetadata.getDataFields(this.getClass());
    for (final Field field : fields) {
      try {
        Object obj = field.get(this);
        if (obj instanceof Key)
          obj = ((Key<?>) obj).copy();
        field.set(copy, obj);
      }
      catch (IllegalArgumentException | IllegalAccessException e) {
        throw new IllegalStateException(e);
      }
    }

    return copy;
  }

  protected void validate()
  {
    // let the impl. put some constraints in here
  }

  protected void validateUpdate(T curr)
  {
    // let the impl. put some constraints in here
  }

  protected void validateDelete(T curr)
  {
    // let the impl. put some constraints in here
  }

  protected void validateThoroughly()
  {
    // let the impl. put some constraints in here
    this.validate();
  }

  protected void onInserted()
  {
    // let the impl. define behavior
  }

  protected void onUpdated(T prev)
  {
    // let the impl. define behavior
  }

  protected void onDeleted(T prev)
  {
    // let the impl. define behavior
  }

  protected void onPostLoad()
  {
    // let the impl. define behavior
  }

  protected void onPreStore()
  {
    // let the impl. define behavior
  }

  protected void onPostStore()
  {
    // let the impl. define behavior
  }

  protected <E extends Enum<E>> void doLoad(
      ResultSet resultset, int columnOffset)
      throws SQLException
  {
    JdbcRecordManagerImpl
        .loadRecordFromResultset(this, resultset, columnOffset);

    this.onPostLoad();
  }

  protected <E extends Enum<E>> int doStore(
      PreparedStatement prep, int columnOffset)
      throws SQLException
  {
    return JdbcRecordManagerImpl.storeRecordIntoPreparedStatement(
        this, prep, columnOffset);
  }

  //

  @Override
  public int hashCode()
  {
    handleZeroId(id);
    return this.id;
  }

  @Override
  public int compareTo(T that)
  {
    handleZeroId(this.id);
    handleZeroId(that.id);

    int diff = this.id - that.id;
    if (diff == 0)
      diff = this.equals(that) ? 0 : -1;
    return diff;
  }

  @Override
  public boolean equals(Object obj)
  {
    handleZeroId(id);

    if (obj == null)
      return false;
    if (obj == this)
      return true;

    if (this.getClass() != obj.getClass())
      return false;

    handleZeroId(((Record<?>) obj).id);
    return (this.id == ((Record<?>) obj).id);
  }

  private static void handleZeroId(int id)
  {
    if (id == 0) {
      throw new IllegalStateException(
          "entity.id is ZERO - trying to put it into a HashSet/HashMap? Don't!");
    }
  }

  public int deepHashcode()
  {
    int hash = 0;
    @SuppressWarnings("unchecked")
    final List<Field> fields =
      RecordTypeMetadata.getDataFields(this.getClass());
    for (final Field field : fields) {
      final Class<?> fieldType = field.getType();
      try {
        if (fieldType == boolean.class) {
          final boolean value = field.getBoolean(this);
          hash ^= value ? 1 : 0;
        }
        else if (fieldType == int.class) {
          final int value = field.getInt(this);
          hash ^= value;
        }
        else if (fieldType == long.class) {
          final long value = field.getLong(this);
          hash ^= (value >> 0);
          hash ^= (value >> 32);
        }
        else if (fieldType == String.class) {
          final String value = (String) field.get(this);
          if (value != null)
            hash ^= value.hashCode();
        }
        else if (fieldType == Key.class) {
          final Key<?> value = (Key<?>) field.get(this);
          if (value != null)
            hash ^= value.hashCode();
        }
        else if (fieldType.isEnum()) {
          final Object enumValue = field.get(this);
          if (enumValue != null)
            hash ^= enumValue.hashCode();
        }
        else {
          throw new UnsupportedOperationException("type: "
              + fieldType.getSimpleName());
        }
      }
      catch (IllegalArgumentException | IllegalAccessException e) {
        throw new IllegalStateException(e);
      }
      hash *= 13;
    }
    return hash;
  }

  public boolean deepEquals(Object obj)
  {
    if (obj == null)
      return false;
    if (obj == this)
      return true;
    if (this.getClass() != obj.getClass())
      return false;
    if (this.id != ((Record<?>) obj).id)
      return false;

    @SuppressWarnings("unchecked")
    final List<Field> fields =
      RecordTypeMetadata.getDataFields(this.getClass());
    for (final Field field : fields) {
      final Class<?> fieldType = field.getType();
      try {
        if (fieldType == boolean.class) {
          final boolean value1 = field.getBoolean(this);
          final boolean value2 = field.getBoolean(obj);
          if (value1 != value2)
            return false;
        }
        else if (fieldType == int.class) {
          final int value1 = field.getInt(this);
          final int value2 = field.getInt(obj);
          if (value1 != value2)
            return false;
        }
        else if (fieldType == long.class) {
          final long value1 = field.getLong(this);
          final long value2 = field.getLong(obj);
          if (value1 != value2)
            return false;
        }
        else if (fieldType == String.class) {
          final String value1 = (String) field.get(this);
          final String value2 = (String) field.get(obj);
          if (!Compare.equals(value1, value2))
            return false;
        }
        else if (fieldType == Key.class) {
          final Key<?> value1 = (Key<?>) field.get(this);
          final Key<?> value2 = (Key<?>) field.get(obj);
          if (!Compare.equals(value1, value2))
            return false;
        }
        else if (fieldType.isEnum()) {
          final Object value1 = field.get(this);
          final Object value2 = field.get(obj);
          if (!Compare.equals(value1, value2))
            return false;
        }
        else {
          throw new UnsupportedOperationException("type: "
              + fieldType.getSimpleName());
        }
      }
      catch (IllegalArgumentException | IllegalAccessException e) {
        throw new IllegalStateException(e);
      }
    }

    return true;
  }

  @Override
  public String toString()
  {
    return this.getClass().getSimpleName() + "["
        + (id == 0 ? "0" : this.getKey()) + "]";
  }

  public String toDetailedString()
  {
    final StringBuilder sb = new StringBuilder();
    sb.append(this.toString()).append("{");

    @SuppressWarnings("unchecked")
    final List<Field> fields =
      RecordTypeMetadata.getDataFields(this.getClass());
    for (final Field field : fields) {
      try {
        if (field != fields.get(0))
          sb.append(", ");
        Object val = field.get(this);
        if (val != null && field.getType().isArray())
          val =
            field.getType().getComponentType().getSimpleName() + "["
                + Array.getLength(val) + "]";
        sb.append(field.getName()).append('=').append(val);
      }
      catch (IllegalArgumentException | IllegalAccessException e) {
        throw new IllegalStateException(e);
      }
    }

    return sb.append("}").toString();
  }
}
