package net.indiespot.sql.data;

import log.Log;
import nl.west.rme.common.util.GenericsUtils;
import nl.west.rme.common.util.MapList;
import nl.west.rme.common.util.WebContextThreadLocal;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Set;


public class RecordManager
{
  private static final Map<String, Class<?>> SIMPLE_NAME_2_TYPE =
    new HashMap<>();

  @SuppressWarnings("unchecked")
  public static <T extends Record<T>> Class<T> lookupTypeByName(
      String simpleName)
  {
    synchronized (SIMPLE_NAME_2_TYPE) {
      Class<?> type = SIMPLE_NAME_2_TYPE.get(simpleName);

      if (type == null && !SIMPLE_NAME_2_TYPE.containsKey(simpleName)) {
        final List<RecordManagerDomain> domains =
          RecordManagerDomain.Access.get();
        for (int i = 0, len = domains.size(); i < len && type == null; i++)
          type = domains.get(i).lookup(simpleName);
        if (type == null)
          throw new IllegalStateException("type not found for: '" + simpleName
              + "'");
        SIMPLE_NAME_2_TYPE.put(simpleName, type);
      }

      return (Class<T>) type;
    }
  }

  private static Set<Class<?>> WARNED_FOR_TYPES = Collections
      .synchronizedSet(new HashSet<Class<?>>());

  public static <T extends Record<T>> T newRecord(Class<T> type)
  {
    final List<RecordManagerDomain> ctxs = RecordManagerDomain.Access.get();
    for (int i = 0, len = ctxs.size(); i < len; i++) {
      final T record = ctxs.get(i).newRecord(type);
      if (record != null)
        return record;
    }

    try {
      if (getThreadLocalConfig().allowOutput)
        if (WARNED_FOR_TYPES.add(type))
          Log.warn("using reflection to create instance of: " + type);
      return type.newInstance();
    }
    catch (InstantiationException | IllegalAccessException e) {
      throw new IllegalStateException(e);
    }
  }

  // ------------------

  public static <T extends Record<T>> Set<Key<T>> fromEntitiesToKeys(
      Set<T> records)
  {
    return fromSetToMap(records).keySet();
  }

  public static <T extends Record<T>> Map<Key<T>, T> fromSetToMap(Set<T> records)
  {
    final Map<Key<T>, T> key2record = new HashMap<>();
    for (final T record : records)
      key2record.put(record.getKey(), record);
    return key2record;
  }

  static Collection<List<Record<?>>> splitOnTypes(Collection<Record<?>> rows)
  {
    final MapList<Class<?>, Record<?>> type2items = new MapList<>();
    for (final Record<?> row : rows)
      type2items.add(row.getClass(), row);
    return type2items.values();
  }

  /**
   * COPY
   */

  public static <T extends Record<T>> List<T> copy(Collection<T> rows)
  {
    final List<T> list = new ArrayList<>();
    for (final T row : rows)
      list.add(row.copy());
    return list;
  }

  public static <T extends Record<T>> String toDetailedString(
      Collection<Record<?>> rows)
  {
    final StringBuilder sb = new StringBuilder();

    for (final List<Record<?>> list : RecordManager.splitOnTypes(rows)) {
      sb.append(list.get(0).getClass().getSimpleName());
      sb.append(" (" + list.size() + "x)\r\n");
      final List<T> typed = GenericsUtils.cast(list);
      for (final T item : typed) {
        sb.append("\t" + item.toDetailedString() + "\r\n");
      }
    }
    return sb.toString();
  }

  public static interface TableDescriptor
  {
    public <T extends Record<T>> String getTableName(Class<T> type);
  }

  public static class Config
  {
    private static final int STACK_SIZE = 16;

    public boolean validateOnStore = true;
    public boolean callEventHandlers = true;
    public boolean allowOutput = true;
    public TableDescriptor tableDescriptor;

    private final boolean[] validateOnStoreStack = new boolean[STACK_SIZE];
    private final boolean[] callEventHandlersStack = new boolean[STACK_SIZE];
    private final boolean[] allowOutputStack = new boolean[STACK_SIZE];
    private final TableDescriptor[] tableDescriptorStack =
      new TableDescriptor[STACK_SIZE];

    private int stackPointer = 0;

    public Config()
    {
      tableDescriptor = new TableDescriptor()
      {
        @Override
        public <T extends Record<T>> String getTableName(Class<T> type)
        {
          return type.getAnnotation(TableName.class).name();
        }
      };
    }

    public void push()
    {
      validateOnStoreStack[stackPointer] = validateOnStore;
      callEventHandlersStack[stackPointer] = callEventHandlers;
      allowOutputStack[stackPointer] = allowOutput;
      tableDescriptorStack[stackPointer] = tableDescriptor;

      stackPointer++;
    }

    public void pop()
    {
      if (stackPointer == 0)
        throw new NoSuchElementException();
      stackPointer--;

      validateOnStore = validateOnStoreStack[stackPointer];
      callEventHandlers = callEventHandlersStack[stackPointer];
      allowOutput = allowOutputStack[stackPointer];
      tableDescriptor = tableDescriptorStack[stackPointer];
    }
  }

  static class ConfigLocal extends WebContextThreadLocal<Config>
  {
    @Override
    protected Config initialValue()
    {
      return new Config();
    }
  };

  static final WebContextThreadLocal<Config> local_config = new ConfigLocal();

  public static Config getThreadLocalConfig()
  {
    return local_config.get();
  }
}
