package nl.west.rme.common.util;

import com.google.gson.JsonArray;
import com.google.gson.JsonPrimitive;
import log.Log;
import java.util.Collection;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.regex.Pattern;


@SuppressWarnings("all")
public class EnumLookup
{
  private static final Map<EnumName, Enum> typename2enum = new HashMap<>();

  public static <T extends Enum<T>> T fromString(Class<T> enumType, String name)
  {
    return fromString(enumType, name, null);
  }

  public static <T extends Enum<T>> T fromString(
      Class<T> enumType, String name, T fallback)
  {
    if (name == null || name.isEmpty())
      return fallback;

    final EnumName typename = new EnumName(enumType, name);
    synchronized (typename2enum) {
      final T enumValue = (T) typename2enum.get(typename);
      if (enumValue != null)
        return (enumValue == FallbackEnum.USE_FALLBACK_VALUE) ? fallback
            : enumValue;
    }

    T found = null;
    {
      name = name.trim().toUpperCase();
      if (!name.isEmpty())
        for (final T value : enumType.getEnumConstants())
          if (value.name().equals(name))
            found = value;
      if (found == null)
        Log.warn("Failed to find enum: " + enumType.getSimpleName() + "."
            + name);
    }

    synchronized (typename2enum) {
      typename2enum.put(typename, (found == null)
          ? FallbackEnum.USE_FALLBACK_VALUE : found);
    }

    return (found == null) ? fallback : found;
  }

  public static <T extends Enum<T>> EnumSet<T> fromStringSet(
      Class<T> enumType, Collection<String> names)
  {
    final EnumSet<T> set = EnumSet.noneOf(enumType);
    for (String name : names) {
      if (name == null)
        continue;

      name = name.trim();
      if (name.isEmpty())
        continue;

      final T value = fromString(enumType, name);
      if (value == null)
        continue;

      set.add(value);
    }
    return set;
  }

  public static <T extends Enum<T>> EnumSet<T> fromStringArray(
      Class<T> enumType, String[] names)
  {
    if (names == null)
      return EnumSet.noneOf(enumType);

    final Set<String> set = new HashSet<>();
    for (final String name : names)
      set.add(name);
    return fromStringSet(enumType, set);
  }

  public static <T extends Enum<T>> EnumSet<T> fromStringCSV(
      Class<T> enumType, String csv, char delim)
  {
    if (csv == null)
      return EnumSet.noneOf(enumType);

    final Set<String> names = new HashSet<>();
    for (final String part : csv.split(Pattern.quote(String.valueOf(delim))))
      names.add(part);
    return fromStringSet(enumType, names);
  }

  public static <T extends Enum<T>> JsonArray toJson(Class<T> enumType)
  {
    final JsonArray json = new JsonArray();
    for (final T enumValue : enumType.getEnumConstants()) {
      json.add(new JsonPrimitive(enumValue.name()));
    }
    return json;
  }

  private static class EnumName
  {
    private final Class type;
    private final String name;

    public EnumName(Class type, String name)
    {
      this.type = type;
      this.name = name;
    }

    @Override
    public int hashCode()
    {
      return type.getName().hashCode() ^ name.hashCode();
    }

    @Override
    public boolean equals(Object obj)
    {
      final EnumName that = (EnumName) obj;
      return (this.type == that.type) && this.name.equals(that.name);
    }
  }

  private static enum FallbackEnum {
    USE_FALLBACK_VALUE;
  }
}
