package nl.west.rme.common.util;

import java.io.ByteArrayOutputStream;
import java.io.Closeable;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;


public class Streams
{

  public static final byte[] readFully(InputStream in)
      throws IOException
  {
    final ByteArrayOutputStream out = new ByteArrayOutputStream();
    copy(in, out);
    return out.toByteArray();
  }

  public static final byte[] readFully(final InputStream in, final int maxSize)
      throws IOException
  {
    final InputStream limitedRangeInputStream = new InputStream()
    {
      private int remaining = maxSize;

      @Override
      public int read()
          throws IOException
      {
        if (remaining <= 0) {
          return -1;
        }
        final int got = in.read();
        remaining--;
        return got;
      }

      @Override
      public int read(byte[] buf)
          throws IOException
      {
        return this.read(buf, 0, buf.length);
      }

      @Override
      public int read(byte[] buf, int off, int len)
          throws IOException
      {
        len = Math.min(len, remaining);
        if (remaining <= 0) {
          return -1;
        }

        final int got = in.read(buf, off, len);
        if (got == -1) {
          return -1;
        }

        remaining -= got;
        return got;
      }
    };

    final ByteArrayOutputStream out = new ByteArrayOutputStream();
    copy(limitedRangeInputStream, out);
    return out.toByteArray();
  }

  //

  public static final long copy(InputStream input, OutputStream output)
      throws IOException
  {
    return copy(input, output, false, null);
  }

  public static final long copy(
      InputStream input, OutputStream output, boolean consumeIOException,
      long[] accum)
      throws IOException
  {
    final byte[] buffer = new byte[4096];

    long copied = 0;

    try {
      while (true) {
        int got;
        try {
          got = input.read(buffer);
        }
        catch (final IOException exc) {
          if (!consumeIOException) {
            throw exc;
          }
          got = -1;
        }
        if (got == -1) {
          break;
        }

        try {
          output.write(buffer, 0, got);
        }
        catch (final IOException exc) {
          if (!consumeIOException)
            throw exc;
          break;
        }

        if (accum != null)
          accum[0] += got;
        else
          copied += got;
      }
    }
    finally {
      try {
        input.close();
      }
      catch (final IOException exc) {
        if (!consumeIOException)
          throw exc;
      }
      finally {
        try {
          output.close();
        }
        catch (final IOException exc) {
          if (!consumeIOException)
            throw exc;
        }
      }
    }

    return copied;
  }

  public static void close(Closeable c)
  {
    close(c, false);
  }

  public static void close(Closeable c, boolean printExc)
  {
    if (c == null) {
      return;
    }

    try {
      c.close();
    }
    catch (final IOException exc) {
      if (printExc) {
        exc.printStackTrace();
      }
    }
  }
}
