package nl.west.rme.common.util;

import log.Log;
import java.util.Calendar;
import java.util.Date;


public class TextDate // "YYYY-MM-DD"
{
  private final Calendar calendar;

  public TextDate(Calendar calendar)
  {
    if (calendar == null)
      throw new NullPointerException();
    this.calendar = calendar;
  }

  public Date snap(Date date, boolean begin)
  {
    if (date == null)
      return null;

    calendar.setTime(date);
    calendar.set(Calendar.HOUR_OF_DAY, begin ? 0 : 23);
    calendar.set(Calendar.MINUTE, begin ? 0 : 59);
    calendar.set(Calendar.SECOND, begin ? 0 : 59);
    calendar.set(Calendar.MILLISECOND, begin ? 0 : 999);

    {
      // sanity checks
      final long millis = calendar.getTime().getTime();
      if (begin && millis % 3600_000L != 0)
        throw new IllegalStateException("date=" + calendar.getTime()
            + ", millis=" + millis);
      if (!begin && millis % 3600_000L != 3600_000L - 1)
        throw new IllegalStateException("date=" + calendar.getTime()
            + ", millis=" + millis);
    }

    calendar.setTime(calendar.getTime());

    {
      // sanity checks
      final long millis = calendar.getTime().getTime();
      if (begin && millis % 3600_000L != 0)
        throw new IllegalStateException("date=" + calendar.getTime()
            + ", millis=" + millis);
      if (!begin && millis % 3600_000L != 3600_000L - 1)
        throw new IllegalStateException("date=" + calendar.getTime()
            + ", millis=" + millis);
    }

    return calendar.getTime();
  }

  public String format(Date date)
  {
    calendar.setTime(date);
    return format(calendar);
  }

  public String format(long timestamp)
  {
    calendar.setTimeInMillis(timestamp);
    return format(//
        calendar.get(Calendar.YEAR), //
        calendar.get(Calendar.MONTH) + 1,//
        calendar.get(Calendar.DAY_OF_MONTH));
  }

  public String now()
  {
    return format(System.currentTimeMillis());
  }

  public String addDays(String date, int days)
  {
    final Calendar cal = parse(date);
    cal.add(Calendar.DATE, days);
    return format(cal);
  }

  public String addYears(String date, int years)
  {
    final Calendar cal = parse(date);
    cal.add(Calendar.YEAR, years);
    return format(cal);
  }

  public int daysBetween(String d1, String d2)
  {
    final Calendar t1 = parse(d1);
    final Calendar t2 = parse(d2);

    final long ms1 = t1.getTimeInMillis();
    final long ms2 = t2.getTimeInMillis();
    final double daysEstimate = (ms2 - ms1) / 86400_000.0D;

    // jump to estimated date
    int days = (int) daysEstimate;
    t1.add(Calendar.DATE, days);

    // correct 1 or 2 days in either direction
    for (int cmp, sig; (cmp = t2.compareTo(t1)) != 0; days += sig)
      t1.add(Calendar.DATE, sig = Integer.signum(cmp));

    return days;
  }

  @SuppressWarnings("unused")
  private void testDaysBetween()
  {
    String d1 = "1999-01-01";
    for (int k = 0; k < 365 * 20; k++) {
      d1 = addDays(d1, 1);
      System.out.println(d1);

      for (int i = -365 * 10; i <= 365 * 10; i++) {
        final String d2 = addDays(d1, i);
        final int ii = daysBetween(d1, d2);
        if (i != ii)
          throw new IllegalStateException();
      }
    }
  }

  public static boolean isLeapYear(int year)
  {
    if (year % 4 != 0)
      return false;
    if (year % 100 != 0)
      return true;
    if (year % 400 != 0)
      return false;
    return true;
  }

  public Calendar parse(String date)
  {
    if (!isValid(date))
      throw new IllegalArgumentException("not a valid date: " + date);

    final String[] parts = date.split("-");
    final int dd = Integer.parseInt(parts[2]);
    final int mm = Integer.parseInt(parts[1]);
    final int yyyy = Integer.parseInt(parts[0]);

    calendar.clear();
    calendar.set(yyyy, mm - 1, dd);
    return (Calendar) calendar.clone();
  }

  public static boolean isValid(String date)
  {
    if (date == null)
      return false;
    final String[] parts = date.split("-");
    if (parts.length != 3)
      return false;

    final int dd = Integer.parseInt(parts[2]);
    final int mm = Integer.parseInt(parts[1]);
    final int yyyy = Integer.parseInt(parts[0]);

    if (dd < 1 || dd > 31)
      return false;
    if (mm < 1 || mm > 12)
      return false;
    if (yyyy < 1900 || yyyy > 2100)
      return false;

    if (mm == 2 && dd > (isLeapYear(yyyy) ? 29 : 28))
      return false;
    if (mm % 2 == (mm / 8) && dd == 31)
      return false;

    return true;
  }

  public static String verify(String date)
  {
    if (!isValid(date))
      throw new IllegalArgumentException("not a valid date: '" + date + "'");
    return date;
  }

  public static String parseDate(String date, String orElse)
  {
    if (isValid(date))
      return date;
    Log.warn("failed to parse date: [" + date + "]");
    return orElse;
  }

  public static String format(int yyyy, int mm, int dd)
  {
    final StringBuilder sb = new StringBuilder();
    sb.append(format(yyyy, 4));
    sb.append('-');
    sb.append(format(mm, 2));
    sb.append('-');
    sb.append(format(dd, 2));
    return sb.toString();
  }

  public static String format(Calendar calendar)
  {
    return format(//
        calendar.get(Calendar.YEAR), //
        calendar.get(Calendar.MONTH) + 1,//
        calendar.get(Calendar.DAY_OF_MONTH));
  }

  public static void apply(String date, Calendar calendar)
  {
    if (!isValid(date))
      throw new IllegalArgumentException("not a valid date: " + date);

    final String[] parts = date.split("-");
    final int dd = Integer.parseInt(parts[2]);
    final int mm = Integer.parseInt(parts[1]);
    final int yyyy = Integer.parseInt(parts[0]);

    calendar.clear();
    calendar.set(yyyy, mm - 1, dd);
  }

  private static String format(int value, int len)
  {
    final StringBuilder sb = new StringBuilder();
    sb.append(value);
    while (sb.length() < len)
      sb.insert(0, '0');
    return sb.toString();
  }
}
